from .memberstate import member_states, MemberState, Threshold
from .vat_check import check_details
from .vies import VIESException, VIESSOAPException, VIESHTTPException, \
     VIESResponseBase, VIESResponse, VIESApproxResponse
from .rates import RateCache
from .vrws import VRWSException, VRWSSOAPException, VRWSHTTPException, \
     VRWSErrorException, Rate, BROADCASTING, TELECOMS, ESERVICES

__all__ = ['member_states', 'MemberState', 'Threshold', 'check_details',
           'VIESException', 'VIESSOAPException', 'VIESHTTPException',
           'VIESResponseBase', 'VIESResponse', 'VIESApproxResponse',
           'RateCache', 'Rates', 'Rate',
           'VRWSException', 'VRWSSOAPException', 'VRWSHTTPException',
           'VRWSErrorException']



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from . import vies, addresscmp

def check_details(vat_number, vat_info={}, requester=None, address_threshold=0.7):
    """Check a VAT number using VIES.  Unlike the functions in
    :py:mod:`vat.vies`, this deals with the fact that different member states
    may behave in different ways, and will always try to do a reasonable job.

    The ``vat_info`` dictionary should contain entries for 'name',
    'street', 'postcode', and 'city'.  It can also contain an entry for
    'state', though that won't be sent to VIES (since VIES doesn't support it).

    This function returns a tuple (match,
    :py:class:`vat.vies.VIESApproxResponse`), where ``match`` is True,
    False or None, which means that we weren't able to determine
    automatically whether or not the details match, but the VAT number itself
    is OK."""

    vat_number = vat_number.upper()
    
    response = vies.check_vat_approx(vat_number, vat_info, requester)

    if not response.valid:
        return (False, response)
    
    # Some member states do fuzzy matching and return these properties,
    # but not all do, so we have to (a) check if they have and
    # (b) check if they're just ignoring everything and not actually
    # checking.
    ms_fuzzy = False
    ms_processed = False
    for m in [ 'name', 'company-type', 'street', 'postcode', 'city' ]:
        status = response.trader_match_info.get(m, '')
        if status:
            ms_fuzzy = True

        if status != vies.MATCH_NOT_PROCESSED:
            ms_processed = True

        if status == vies.MATCH_INVALID:
            return (False, response)

    # Ignore member states if they don't actually check anything
    if ms_fuzzy and not ms_processed:
        ms_fuzzy = False

    if not ms_fuzzy:
        if vat_number.startswith('DE'):
            # Germany IS ANNOYING!
            # The Germans are not doing fuzzy matching *AND* they are also
            # not returning data (they just pass it straight back unchanged).
            # This is against EU law, but what do they care?
            vies_address = ''
        else:
            # Member states that don't do fuzzy matching, are *supposed* to
            # return the trader's address so that we can do it ourselves.
            vies_address = response.trader_info.get('address', '')
            if not vies_address:
                vies_address = []
                for m in [ 'name', 'street', 'postcode', 'city' ]:
                    item = response.trader_info.get(m, '')
                    if item:
                        vies_address.append(item)
                vies_address = ' '.join(vies_address)

        if not vies_address:
            return (None, response)

        user_address = []
        for item in ['street', 'postcode', 'city', 'state']:
            info = vat_info.get(item, None)
            if info:
                user_address.append(info)
        user_address = ' '.join(user_address)

        score = addresscmp.compare(vies_address, user_address)
        if score < address_threshold:
            return (False, response)

    return (True, response)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import re
import urllib
import datetime
import xml.sax.saxutils
import time
import six
from six.moves import http_client
from dateutil import tz
from lxml import etree

VIES_HOST = str('ec.europa.eu')
VIES_PATH = str('/taxation_customs/vies/services/checkVatService')

class VIESException(Exception):
    pass

_vies_fault_re = re.compile(r'/(INVALID_INPUT|SERVICE_UNAVAILABLE|MS_UNAVAILABLE|TIMEOUT|SERVER_BUSY)/')

class VIESSOAPException(VIESException):
    def __init__(self, code, string, actor, detail):
        self.code = code
        self.string = string
        self.actor = actor
        self.detail = detail

        m = _vies_fault_re.search(string)
        if m:
            self.fault_type = m.group(0)
        else:
            self.fault_type = 'UNKNOWN'
        
    def __repr__(self):
        return 'VIESSOAPException(%r, %r, %r, %r)' % (self.code,
                                                      self.string,
                                                      self.actor,
                                                      self.detail)

    def __unicode__(self):
        return '%s - %s (%s)\n%s' % (self.code, self.string, self.actor, self.detail)

    def __str__(self):
        return str(self.__unicode__())
    
class VIESHTTPException(VIESException):
    def __init__(self, code, message):
        self.code = code
        self.message = message

    def __repr__(self):
        return 'VIESHTTPException(%r, %r)' % (self.code, self.message)

    def __unicode__(self):
        return '%s - %s' % (self.code, self.message)

    def __str__(self):
        return str(self.__unicode__())
    
class VIESResponseBase(object):
    def __init__(self, country, vat_number, request_date, valid):
        self.country = country
        self.vat_number = vat_number
        self.request_date = request_date
        self.valid = valid
    
class VIESResponse(VIESResponseBase):
    def __init__(self, country, vat_number, request_date,
                 valid, name, address):
        super(VIESResponse, self).__init__(country, vat_number,
                                           request_date, valid)
        self.name = name
        self.address = address

    def __repr__(self):
        return 'VIESResponse(%r, %r, %r, %r, %r, %r)' % (self.country,
                                                         self.vat_number,
                                                         self.request_date,
                                                         self.valid,
                                                         self.name,
                                                         self.address)
    
class VIESApproxResponse(VIESResponseBase):
    def __init__(self, country, vat_number, request_date, valid,
                 trader_info, trader_match_info, request_id):
        super(VIESApproxResponse, self).__init__(country, vat_number,
                                                 request_date, valid)
        self.trader_info = trader_info
        self.trader_match_info = trader_match_info
        self.request_id = request_id
        
    def __repr__(self):
        return 'VIESApproxResponse(%r, %r, %r, %r, %r, %r, %r)' \
          % (self.country, self.vat_number, self.request_date, self.valid,
             self.trader_info, self.trader_match_info, self.request_id)

_strip_re = re.compile(r'[^A-Za-z0-9]+')
def _strip_vat(vn):
    """Remove any whitespace and punctuation from a VAT number."""
    return _strip_re.sub('', vn)

SOAP_NS = '{http://schemas.xmlsoap.org/soap/envelope/}'
VIES_NS = '{urn:ec.europa.eu:taxud:vies:services:checkVat:types}'

_date_re = re.compile(r'(?P<year>[0-9]{4})-(?P<month>[0-9]{2})-(?P<day>[0-9]{2})(?:Z|(?P<tzsign>[-+])(?P<tzhours>[0-9]{2}):(?P<tzmins>[0-9]{2}))?$')
def _parse_date(d):
    m = _date_re.match(d)
    if not m:
        return None
    year = int(m.group('year'))
    month = int(m.group('month'))
    day = int(m.group('day'))
    tzsign = m.group('tzsign')
    tzinfo = None
    if tzsign is not None:
        tzhours = int(m.group('tzhours'))
        tzmins = int(m.group('tzmins'))
        tzoffset = (tzhours * 60 + tzmins) * 60
        if tzsign == '-':
            tzoffset = -tzoffset
        tzinfo = tz.tzoffset(None, tzoffset)
    elif d.endswith('Z'):
        tzinfo = tz.tzutc()
    return datetime.datetime(year, month, day, tzinfo=tzinfo)

def _parse_boolean(b):
    if b == 'true' or b == '1':
        return True
    return False

def check_vat(vat_number):
    """Check a VAT number using VIES.  Returns a VIESResponse object on
    success, or in case of error raises an exception."""
    vat_number = _strip_vat(vat_number)
    country_code = vat_number[:2]
    number = vat_number[2:]

    message = '''<?xml version="1.0" encoding="UTF-8" ?>
<env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/"
 env:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
  <env:Body xmlns:vies="urn:ec.europa.eu:taxud:vies:services:checkVat:types">
    <vies:checkVat>
      <vies:countryCode>%s</vies:countryCode>
      <vies:vatNumber>%s</vies:vatNumber>
    </vies:checkVat>
  </env:Body>
</env:Envelope>''' % (country_code, number)
    message = message.encode('utf-8')
    
    headers = { b'Content-type': b'text/xml',
                b'SOAPAction': b'urn:ec.europa.eu:taxud:vies:services:checkVat' }

    tries = 0
    response = None
    while response is None or (tries < 5
                               and response.status >= 500
                               and response.status <= 599):
        if tries > 0:
            time.sleep(tries)
        tries += 1
        conn = http_client.HTTPConnection(VIES_HOST)
        conn.request(str('POST'), VIES_PATH, message, headers)
        response = conn.getresponse()

    if response.status != 200:
        raise VIESHTTPException(response.status, response.reason)

    tree = etree.parse(response)
    root = tree.getroot()

    if root.tag.lower() == SOAP_NS + 'fault':
        faultcode = root.find('./' + SOAP_NS + 'faultcode').text
        faultstring = root.find('./' + SOAP_NS + 'faultstring').text
        faultactor = root.find('./' + SOAP_NS + 'faultactor').text
        detail = root.find('./' + SOAP_NS + 'detail').text
        raise VIESSOAPException(faultcode, faultstring, faultactor, detail)
    
    if root.tag.lower() != SOAP_NS + 'envelope':
        raise ValueError('Bad SOAP reply "%s"' % etree.tostring(tree))

    resp = root.find('./' + SOAP_NS + 'Body/' + VIES_NS + 'checkVatResponse')

    if resp is None:
        raise ValueError('Bad SOAP reply "%s"' % etree.tostring(tree))
    
    country_code = resp.find('./' + VIES_NS + 'countryCode').text
    number = resp.find('./' + VIES_NS + 'vatNumber').text
    request_date = _parse_date(resp.find('./' + VIES_NS + 'requestDate').text)
    valid = _parse_boolean(resp.find('./' + VIES_NS + 'valid').text)
    name = resp.find('./' + VIES_NS + 'name').text
    address = resp.find('./' + VIES_NS + 'address').text

    if name == '---':
        name = None
    if address == '---':
        address = None
    
    return VIESResponse(country_code, number, request_date, valid, name, address)

_eltnames = {
    'name': 'traderName',
    'company-type': 'traderCompanyType',
    'street': 'traderStreet',
    'postcode': 'traderPostcode',
    'city': 'traderCity'
    }

_respeltnames = {
    'traderName': 'name',
    'traderCompanyType': 'company-type',
    'traderAddress': 'address',
    'traderStreet': 'street',
    'traderPostcode': 'postcode',
    'traderCity': 'city'
    }

_respmatchnames = {
    'traderNameMatch': 'name',
    'traderCompanyTypeMatch': 'company-type',
    'traderStreetMatch': 'street',
    'traderPostcodeMatch': 'postcode',
    'traderCityMatch': 'city',
    }

MATCH_VALID = 'valid'
MATCH_INVALID = 'invalid'
MATCH_NOT_PROCESSED = 'not-processed'

_matchcodes = {
    1: MATCH_VALID,
    2: MATCH_INVALID,
    3: MATCH_NOT_PROCESSED
    }

def check_vat_approx(vat_number, extra={}, requester=None):
    """Check a VAT number using VIES, passing in additional information about
    the entity being checked.  Returns a VIESApproxResponse object on
    success, or in case of error raises an exception.

    The keys in the `extra` dictionary are as follows:

      name
      company-type
      street
      postcode
      city

    All keys are optional.

    You can also pass in the VAT number of the requesting entity; this too
    is optional."""

    vat_number = _strip_vat(vat_number)
    country_code = vat_number[:2]
    number = vat_number[2:]

    extra_tags = []
    for k,v in six.iteritems(extra):
        t = _eltnames.get(k, None)
        if t:
            v = xml.sax.saxutils.escape(v, { "'": '&#x2019;' })
            extra_tags.append('<vies:%s>%s</vies:%s>' % (t, v, t))

    if requester:
        requester = _strip_vat(requester)
        req_cc = requester[:2]
        req_num = requester[2:]
        extra_tags.append('<vies:requesterCountryCode>%s</vies:requesterCountryCode><vies:requesterVatNumber>%s</vies:requesterVatNumber>' % (req_cc, req_num))
            
    message = '''<?xml version="1.0" encoding="UTF-8" ?>
<env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/"
 env:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
  <env:Body xmlns:vies="urn:ec.europa.eu:taxud:vies:services:checkVat:types">
    <vies:checkVatApprox>
      <vies:countryCode>%s</vies:countryCode>
      <vies:vatNumber>%s</vies:vatNumber>%s
    </vies:checkVatApprox>
  </env:Body>
</env:Envelope>''' % (country_code, number, '\n'.join(extra_tags))
    message = message.encode('utf-8')

    headers = { b'Content-type': b'text/xml',
                b'SOAPAction': b'urn:ec.europa.eu:taxud:vies:services:checkVatApprox' }

    tries = 0
    response = None
    while response is None or (tries < 5
                               and response.status >= 500
                               and response.status <= 599):
        if tries > 0:
            time.sleep(tries)
        tries += 1
        conn = http_client.HTTPConnection(VIES_HOST)
        conn.request(str('POST'), VIES_PATH, message, headers)
        response = conn.getresponse()
    
    if response.status != 200:
        raise VIESHTTPException(response.status, response.reason)

    tree = etree.parse(response)
    root = tree.getroot()

    if root.tag.lower() == SOAP_NS + 'fault':
        faultcode = root.find('./' + SOAP_NS + 'faultcode').text
        faultstring = root.find('./' + SOAP_NS + 'faultstring').text
        faultactor = root.find('./' + SOAP_NS + 'faultactor').text
        detail = root.find('./' + SOAP_NS + 'detail').text
        raise VIESSOAPException(faultcode, faultstring, faultactor, detail)
    
    if root.tag.lower() != SOAP_NS + 'envelope':
        raise ValueError('Bad SOAP reply "%s"' % etree.tostring(tree))

    resp = root.find('./' + SOAP_NS + 'Body/' + VIES_NS
                     + 'checkVatApproxResponse')

    if resp is None:
        raise ValueError('Bad SOAP reply "%s"' % etree.tostring(tree))

    country_code = resp.find('./' + VIES_NS + 'countryCode').text
    number = resp.find('./' + VIES_NS + 'vatNumber').text
    request_date = _parse_date(resp.find('./' + VIES_NS + 'requestDate').text)
    valid = _parse_boolean(resp.find('./' + VIES_NS + 'valid').text)
    request_id = resp.find('./' + VIES_NS + 'requestIdentifier').text
    
    info = {}
    for t,k in six.iteritems(_respeltnames):
        elt = resp.find('./' + VIES_NS + t)
        if elt is not None:
            if elt.text == '---':
                info[k] = None
            else:
                info[k] = elt.text

    match = {}
    for t,k in six.iteritems(_respmatchnames):
        elt = resp.find('./' + VIES_NS + t)
        if elt is not None:
            v = int(elt.text)
            match[k] = _matchcodes.get(v, '<unknown %d>' % v)

    return VIESApproxResponse(country_code, number, request_date, valid,
                              info, match, request_id)



display_name = "Metafone"
library_name = "metafone"
version = "0.5"
author = "Alastair Houghton"
author_email = "alastair@alastairs-place.net"
license = "BSD"
url = "https://github.com/al45tair/metaphone"
description = "A Python implementation of the double metaphone algorithms."



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import datetime
from decimal import Decimal as D

from . import vrws

class RateCache (object):
    """Manages a cache of VAT rates fetched from europa.eu's
       VATRateWebService.  Most queries will come from the cache, which will
       only fetch new rates once per day per member state."""
    
    # Historic reduced rates - VRWS currently isn't supplying reduced rate
    # information
    
    # I've tried to pick, as the first entry (with no detail) the most likely
    # rate you'll want.  But you MUST check.  Also, as of 1st Sept 2015, NONE
    # of this information is available via the web service, so the detail may
    # change to something else.
    historic_reduced_rates = {
        'BE': (D('6.0'), (D('12.0'), 'Alternate'), (D('12.0'), 'Parking')),
        'BG': (D('9.0'),),
        'CZ': (D('15.0'), (D('10.0'), 'Alternate')),
        'DK': None,
        'DE': (D('7.0'),),
        'EE': (D('9.0'),),
        'EL': (D('13.0'), (D('6.0'), 'Alternate')),
        'ES': (D('10.0'), (D('4.0'), 'Super Reduced')),
        'FR': (D('5.5'), (D('10.0'), 'Alternate'), (D('2.1'), 'Super Reduced')),
        'HR': (D('5.0'), (D('13.0'), 'Alternate')),
        'IE': (D('13.5'), (D('9.0'), 'Alternate'), (D('4.8'), 'Super Reduced'),
               (D('13.5'), 'Parking')),
        'IT': (D('10.0'), (D('4.0'), 'Super Reduced')),
        'CY': (D('5.0'), (D('9.0'), 'Alternate')),
        'LV': (D('12.0'),),
        'LT': (D('9.0'), (D('5.0'), 'Alternate')),
        'LU': (D('8.0'), (D('3.0'), 'Super Reduced'), (D('14.0'), 'Parking')),
        'HU': (D('5.0'), (D('18.0'), 'Alternate')),
        'MT': (D('5.0'), (D('7.0'), 'Alternate')),
        'NL': (D('6.0'),),
        'AT': (D('10.0'), (D('12.0'), 'Parking')),
        'PL': (D('8.0'), (D('5.0'), 'Alternate')),
        'PT': (D('6.0'), (D('13.0'), 'Alternate'), (D('13.0'), 'Parking')),
        'RO': (D('9.0'), (D('5.0'), 'Alternate')),
        'SI': (D('9.5'),),
        'SK': (D('10.0'),),
        'FI': (D('10.0'), (D('14.0'), 'Alternate')),
        'SE': (D('6.0'), (D('12.0'), 'Alternate')),
        'GB': (D('5.0'),),
        }
        
    def __init__(self):
        self.rates = {}

    def _get_rates(self, member_state):
        if not isinstance(member_state, basestring):
            member_state = member_state.code
        today = datetime.date.today()
        rinfo = self.rates.get(member_state, None)
        if rinfo is not None:
            if rinfo[1] == today:
                return rinfo[0]
        rates = vrws.get_rates(member_state, date=today)
        self.rates[member_state] = (rates, today)
        return rates

    def regions(self, member_state):
        """Retrieve a list of regions for the given member state."""
        rates = self._get_rates(member_state)
        return rates.regions.keys()

    def _best_rate(self, rates):
        today = datetime.date.today()
        rate_date = None
        best_rate = None
        for rate in rates:
            if rate.application_date > today or rate.detail is not None:
                continue
            if rate_date is None or rate.application_date > rate_date:
                best_rate = rate
        return best_rate
    
    def standard_rates(self, member_state, region=None):
        """Retrieve the set of standard rates for the given member state,
        optionally for the specified region.

        N.B.: This method returns a LIST of rates, and you should use the
        date and detail information to find the appropriate one.  There may
        be multiple rates with different application dates, *and* there may
        be rates with detail specifications in the list."""
        rates = self._get_rates(member_state)
        if region is not None:
            rates = rates.regions[region]
        return rates.types.get(vrws.STANDARD, [])

    def standard_rate(self, member_state, region=None):
        """Return today’s ordinary standard rate for the given member state
        and (optional) region."""
        return self._best_rate(self.standard_rates(member_state, region))

    def _to_rate(self, rate_info, the_date):
        if isinstance(rate_info, tuple):
            return vrws.Rate(rate_info[0], the_date, rate_info[1])
        else:
            return vrws.Rate(rate_info, the_date, None)
    
    def reduced_rates(self, member_state, region=None):
        """Retrieve the set of reduced rates for the given member state,
        optionally for the specified region.

        N.B.: This method returns a LIST of rates, and you should use the
        date and detail information to find the appropriate one.  There may
        be multiple rates with different application dates, *and* there may
        be rates with detail specifications in the list.

        N.B. 2: As of 1st September 2015, most member states are not providing
        reduced rate information via the web service.  If we find that a
        member state has not provided the necessary information, we fall back
        to a built-in table."""
        rates = self._get_rates(member_state)
        if region is not None:
            rates = rates.regions[region]
        rates = rates.types.get(vrws.REDUCED, None)
        if rates is None:
            rates = self.historic_reduced_rates.get(member_state, None)
            if rates is None:
                return []
            the_date = datetime.date(2015,9,1)
            rates = [self._to_rate(r, the_date) for r in rates]
        return rates
    
    def reduced_rate(self, member_state, region=None):
        """Return today’s ordinary reduced rate for the given member state
        and (optional) region."""
        return self._best_rate(self.reduced_rates(member_state, region))
    
    def category_rates(self, member_state, category, region=None):
        """Retrieve the set of rates for the given member state, category
        and (optional) region.

        N.B.: This method returns a LIST of rates, and you should use the
        date and detail information to find the appropriate one.  There may
        be multiple rates with different application dates, *and* there may
        be rates with detail specifications in the list."""
        rates = self._get_rates(member_state)
        if region is not None:
            rates = rates.regions[region]
        return rates.categories.get(category, [])
    
    def category_rate(self, member_state, category, region=None):
        """Return today’s rate for the given member state, category and
        (optional) region."""
        return self._best_rate(self.category_rates(member_state,
                                                   category,
                                                   region))
    
    def categories(self, member_state, region=None):
        """Return a list of categories for the given member state and
        (optional) region."""
        rates = self._get_rates(member_state)
        if region is not None:
            rates = rates.regions[region]
        return rates.categories.keys()



# -*- coding: utf-8 -*-
"""
The original Metaphone algorithm was published in 1990 as an improvement over
the Soundex algorithm. Like Soundex, it was limited to English-only use. The
Metaphone algorithm does not produce phonetic representations of an input word
or name; rather, the output is an intentionally approximate phonetic
representation. The approximate encoding is necessary to account for the way
speakers vary their pronunciations and misspell or otherwise vary words and
names they are trying to spell.

The Double Metaphone phonetic encoding algorithm is the second generation of
the Metaphone algorithm. Its implementation was described in the June 2000
issue of C/C++ Users Journal. It makes a number of fundamental design
improvements over the original Metaphone algorithm.

It is called "Double" because it can return both a primary and a secondary code
for a string; this accounts for some ambiguous cases as well as for multiple
variants of surnames with common ancestry. For example, encoding the name
"Smith" yields a primary code of SM0 and a secondary code of XMT, while the
name "Schmidt" yields a primary code of XMT and a secondary code of SMT--both
have XMT in common.

Double Metaphone tries to account for myriad irregularities in English of
Slavic, Germanic, Celtic, Greek, French, Italian, Spanish, Chinese, and other
origin. Thus it uses a much more complex ruleset for coding than its
predecessor; for example, it tests for approximately 100 different contexts of
the use of the letter C alone.

This script implements the Double Metaphone algorithm (c) 1998, 1999 originally
implemented by Lawrence Philips in C++. It was further modified in C++ by Kevin
Atkinson (http://aspell.net/metaphone/). It was translated to C by Maurice
Aubrey <maurice@hevanet.com> for use in a Perl extension. A Python version was
created by Andrew Collins on January 12, 2007, using the C source
(http://www.atomodo.com/code/double-metaphone/metaphone.py/view).

  Updated 2007-02-14 - Found a typo in the 'gh' section (0.1.1)
  Updated 2007-12-17 - Bugs fixed in 'S', 'Z', and 'J' sections (0.2;
                       Chris Leong)
  Updated 2009-03-05 - Various bug fixes against the reference C++
                       implementation (0.3; Matthew Somerville)
  Updated 2012-07    - Fixed long lines, added more docs, changed names,
                       reformulated as objects, fixed a bug in 'G'
                       (0.4; Duncan McGreggor)
  Updated 2013-06    - Enforced unicode literals (0.5; Ian Beaver)
"""
from __future__ import unicode_literals
from .word import Word


VOWELS = ['A', 'E', 'I', 'O', 'U', 'Y']
SILENT_STARTERS = ["GN", "KN", "PN", "WR", "PS"]


class DoubleMetaphone(object):
    """
    """
    def __init__(self):
        self.position = 0
        self.primary_phone = ""
        self.secondary_phone = ""
        # next is used set to a tuple of the next characters in the primary and
        # secondary codes and to indicate how many characters to move forward
        # in the string.  The secondary code letter is given only when it is
        # different than the primary. This is just a trick to make the code
        # easier to write and read. The default action is to add nothing and
        # move to next char.
        self.next = (None, 1)

    def check_word_start(self):
        # skip these silent letters when at start of word
        if self.word.get_letters(0, 2) in SILENT_STARTERS:
            self.position += 1
        # Initial 'X' is pronounced 'Z' e.g. 'Xavier'
        if self.word.get_letters(0) == 'X':
            # 'Z' maps to 'S'
            self.primary_phone = self.secondary_phone = 'S'
            self.position += 1

    def process_initial_vowels(self):
        # XXX do we need this next set? it should already be done...
        self.next = (None, 1)
        # all init vowels now map to 'A'
        if self.position == self.word.start_index:
            self.next = ('A', 1)

    def process_b(self):
        # "-mb", e.g., "dumb", already skipped over... see 'M' below
        if self.word.buffer[self.position + 1] == 'B':
            self.next = ('P', 2)
        else:
            self.next = ('P', 1)

    def process_c(self):
        buffer = self.word.buffer
        position = self.position
        start_index = self.word.start_index
        # various germanic
        if (position > start_index + 1
            and buffer[position - 2] not in VOWELS
            and buffer[position - 1:self.position + 2] == 'ACH'
            and buffer[position + 2] not in ['I']
            and (buffer[position + 2] not in ['E']
                 or buffer[position - 2:position + 4] in [
                    'BACHER', 'MACHER'])):
            self.next = ('K', 2)
        # special case 'CAESAR'
        elif (position == start_index
              and buffer[start_index:start_index + 6] == 'CAESAR'):
            self.next = ('S', 2)
        # italian 'chianti'
        elif buffer[position:position + 4] == 'CHIA':
            self.next = ('K', 2)
        elif buffer[position:position + 2] == 'CH':
            # find 'michael'
            if (position > start_index
                and buffer[position:position + 4] == 'CHAE'):
                self.next = ('K', 'X', 2)
            elif (position == start_index
                  and (buffer[position + 1:position + 6] in ['HARAC', 'HARIS']
                  or buffer[position + 1:position + 4] in ["HOR", "HYM", "HIA",
                                                           "HEM"])
                  and buffer[start_index:start_index + 5] != 'CHORE'):
                self.next = ('K', 2)
            # germanic, greek, or otherwise 'ch' for 'kh' sound
            elif (
                buffer[start_index:start_index + 4] in ['VAN ', 'VON ']
                or buffer[start_index:start_index + 3] == 'SCH'
                or buffer[position - 2:position + 4] in ["ORCHES", "ARCHIT",
                                                         "ORCHID"]
                or buffer[position + 2] in ['T', 'S']
                or (
                    (buffer[position - 1] in ["A", "O", "U", "E"]
                     or position == start_index)
                    and (buffer[position + 2] in [
                        "L", "R", "N", "M", "B", "H", "F", "V", "W"]))):
                self.next = ('K', 2)
            else:
                if position > start_index:
                    if buffer[start_index:start_index + 2] == 'MC':
                        self.next = ('K', 2)
                    else:
                        self.next = ('X', 'K', 2)
                else:
                    self.next = ('X', 2)
        # e.g, 'czerny'
        elif (buffer[position:position + 2] == 'CZ'
              and buffer[position - 2:position + 2] != 'WICZ'):
            self.next = ('S', 'X', 2)
        # e.g., 'focaccia'
        elif buffer[position + 1:position + 4] == 'CIA':
            self.next = ('X', 3)
        # double 'C', but not if e.g. 'McClellan'
        elif (
            buffer[position:position + 2] == 'CC'
            and not (position == (start_index + 1)
                     and buffer[start_index] == 'M')):
            #'bellocchio' but not 'bacchus'
            if (buffer[position + 2] in ["I", "E", "H"]
                and buffer[position + 2:position + 4] != 'HU'):
                # 'accident', 'accede' 'succeed'
                if (
                    (position == (start_index + 1)
                     and buffer[start_index] == 'A')
                    or buffer[position - 1:position + 4] in [
                        'UCCEE', 'UCCES']):
                    self.next = ('KS', 3)
                # 'bacci', 'bertucci', other italian
                else:
                    self.next = ('X', 3)
            else:
                self.next = ('K', 2)
        elif buffer[position:position + 2] in ["CK", "CG", "CQ"]:
            self.next = ('K', 2)
        elif buffer[position:position + 2] in ["CI", "CE", "CY"]:
            # italian vs. english
            if buffer[position:position + 3] in ["CIO", "CIE", "CIA"]:
                self.next = ('S', 'X', 2)
            else:
                self.next = ('S', 2)
        else:
            # name sent in 'mac caffrey', 'mac gregor'
            if buffer[position + 1:position + 3] in [" C", " Q", " G"]:
                self.next = ('K', 3)
            else:
                if (buffer[position + 1] in ["C", "K", "Q"]
                    and buffer[position + 1:position + 3] not in ["CE", "CI"]):
                    self.next = ('K', 2)
                # default for 'C'
                else:
                    self.next = ('K', 1)

    def process_d(self):
        if self.word.buffer[self.position:self.position + 2] == 'DG':
            # e.g. 'edge'
            if self.word.buffer[self.position + 2] in ['I', 'E', 'Y']:
                self.next = ('J', 3)
            else:
                self.next = ('TK', 2)
        elif self.word.buffer[self.position:self.position + 2] in ['DT', 'DD']:
            self.next = ('T', 2)
        else:
            self.next = ('T', 1)

    def process_f(self):
        if self.word.buffer[self.position + 1] == 'F':
            self.next = ('F', 2)
        else:
            self.next = ('F', 1)

    def process_g(self):
        buffer = self.word.buffer
        position = self.position
        start_index = self.word.start_index
        if buffer[position + 1] == 'H':
            if (position > start_index
                and buffer[position - 1] not in VOWELS):
                self.next = ('K', 2)
            elif position < (start_index + 3):
                # 'ghislane', ghiradelli
                if position == start_index:
                    if buffer[position + 2] == 'I':
                        self.next = ('J', 2)
                    else:
                        self.next = ('K', 2)
            # Parker's rule (with some further refinements) - e.g., 'hugh'
            elif (
                (position > (start_index + 1)
                 and buffer[position - 2] in ['B', 'H', 'D'])
                or (position > (start_index + 2)
                 and buffer[position - 3] in ['B', 'H', 'D'])
                or (position > (start_index + 3)
                 and buffer[position - 4] in ['B', 'H'])):
                self.next = (None, 2)
            else:
                # e.g., 'laugh', 'McLaughlin', 'cough', 'gough', 'rough',
                # 'tough'
                if (position > (start_index + 2)
                    and buffer[position - 1] == 'U'
                    and buffer[position - 3] in [
                        "C", "G", "L", "R", "T"]):
                    self.next = ('F', 2)
                else:
                    if (position > start_index
                        and buffer[position - 1] != 'I'):
                        self.next = ('K', 2)
        elif buffer[position + 1] == 'N':
            if (position == (start_index + 1)
                and buffer[start_index] in VOWELS
                and not self.word.is_slavo_germanic):
                self.next = ('KN', 'N', 2)
            else:
                # not e.g. 'cagney'
                if (buffer[position + 2:position + 4] != 'EY'
                    and buffer[position + 1] != 'Y'
                    and not self.word.is_slavo_germanic):
                    self.next = ('N', 'KN', 2)
                else:
                    self.next = ('KN', 2)
        # 'tagliaro'
        elif (buffer[position + 1:position + 3] == 'LI'
              and not self.word.is_slavo_germanic):
            self.next = ('KL', 'L', 2)
        # -ges-,-gep-,-gel-, -gie- at beginning
        elif (position == start_index
              and (buffer[position + 1] == 'Y'
              or buffer[position + 1:position + 3] in [
                "ES", "EP", "EB", "EL", "EY", "IB", "IL", "IN", "IE", "EI",
                "ER"])):
            self.next = ('K', 'J', 2)
        # -ger-,  -gy-
        elif (
            (buffer[position + 1:position + 3] == 'ER'
             or buffer[position + 1] == 'Y')
            and buffer[start_index:start_index + 6] not in [
                "DANGER", "RANGER", "MANGER"]
            and buffer[position - 1] not in ['E', 'I']
            and buffer[position - 1:position + 2] not in ['RGY', 'OGY']):
            self.next = ('K', 'J', 2)
        # italian e.g, 'biaggi'
        elif (
            buffer[position + 1] in ['E', 'I', 'Y']
            or buffer[position - 1:position + 3] in [
                "AGGI", "OGGI"]):
            # obvious germanic
            if (buffer[start_index:start_index + 4] in ['VON ', 'VAN ']
                or buffer[start_index:start_index + 3] == 'SCH'
                or buffer[position + 1:position + 3] == 'ET'):
                self.next = ('K', 2)
            else:
                # always soft if french ending
                if buffer[position + 1:position + 5] == 'IER ':
                    self.next = ('J', 2)
                else:
                    self.next = ('J', 'K', 2)
        elif buffer[position + 1] == 'G':
            self.next = ('K', 2)
        else:
            self.next = ('K', 1)

    def process_h(self):
        # only keep if self.word.start_index & before vowel or btw. 2 vowels
        if ((self.position == self.word.start_index
             or self.word.buffer[self.position - 1] in VOWELS)
            and self.word.buffer[self.position + 1] in VOWELS):
            self.next = ('H', 2)
        # (also takes care of 'HH')
        else:
            self.next = (None, 1)

    def process_j(self):
        buffer = self.word.buffer
        position = self.position
        start_index = self.word.start_index
        # obvious spanish, 'jose', 'san jacinto'
        if (buffer[self.position:self.position + 4] == 'JOSE'
            or buffer[start_index:start_index + 4] == 'SAN '):
            if (
                (position == start_index and buffer[position + 4] == ' ')
                or buffer[start_index:start_index + 4] == 'SAN '):
                self.next = ('H', )
            else:
                self.next = ('J', 'H')
        # Yankelovich/Jankelowicz
        elif (position == start_index
              and buffer[self.position:self.position + 4] != 'JOSE'):
            self.next = ('J', 'A')
        else:
            # spanish pron. of e.g. 'bajador'
            if (buffer[position - 1] in VOWELS
                and not self.word.is_slavo_germanic
                and buffer[position + 1] in ['A', 'O']):
                self.next = ('J', 'H')
            else:
                if position == self.word.end_index:
                    self.next = ('J', ' ')
                else:
                    if (buffer[position + 1] not in ["L", "T", "K", "S", "N",
                                               "M", "B", "Z"]
                        and buffer[position - 1] not in ["S", "K", "L"]):
                        self.next = ('J',)
                    else:
                        self.next = (None, )
        if buffer[position + 1] == 'J':
            self.next = self.next + (2,)
        else:
            self.next = self.next + (1,)

    def process_k(self):
        if self.word.buffer[self.position + 1] == 'K':
            self.next = ('K', 2)
        else:
            self.next = ('K', 1)

    def process_l(self):
        buffer = self.word.buffer
        position = self.position
        end_index = self.word.end_index
        if buffer[position + 1] == 'L':
            # spanish e.g. 'cabrillo', 'gallegos'
            if ((position == (end_index - 2)
                 and buffer[position - 1:position + 3] in [
                    "ILLO", "ILLA", "ALLE"])
                or ((buffer[end_index - 1:end_index + 1] in ["AS", "OS"]
                     or buffer[end_index] in ["A", "O"])
                    and buffer[position - 1:position + 3] == 'ALLE')):
                self.next = ('L', '', 2)
            else:
                self.next = ('L', 2)
        else:
            self.next = ('L', 1)

    def process_m(self):
        buffer = self.word.buffer
        position = self.position
        if ((buffer[position + 1:position + 4] == 'UMB'
             and (position + 1 == self.word.end_index
                  or buffer[position + 2:position + 4] == 'ER'))
            or buffer[position + 1] == 'M'):
            self.next = ('M', 2)
        else:
            self.next = ('M', 1)

    def process_n(self):
        if self.word.buffer[self.position + 1] == 'N':
            self.next = ('N', 2)
        else:
            self.next = ('N', 1)

    def process_p(self):
        if self.word.buffer[self.position + 1] == 'H':
            self.next = ('F', 2)
        # also account for "campbell", "raspberry"
        elif self.word.buffer[self.position + 1] in ['P', 'B']:
            self.next = ('P', 2)
        else:
            self.next = ('P', 1)

    def process_q(self):
        if self.word.buffer[self.position + 1] == 'Q':
            self.next = ('K', 2)
        else:
            self.next = ('K', 1)

    def process_r(self):
        buffer = self.word.buffer
        position = self.position
        end_index = self.word.end_index
        # french e.g. 'rogier', but exclude 'hochmeier'
        if (position == end_index
            and not self.word.is_slavo_germanic
            and buffer[position - 2:position] == 'IE'
            and buffer[position - 4:position - 2] not in ['ME', 'MA']):
            self.next = ('', 'R')
        else:
            self.next = ('R',)
        if buffer[position + 1] == 'R':
            self.next = self.next + (2,)
        else:
            self.next = self.next + (1,)

    def process_s(self):
        buffer = self.word.buffer
        position = self.position
        start_index = self.word.start_index
        end_index = self.word.end_index
        # special cases 'island', 'isle', 'carlisle', 'carlysle'
        if buffer[position - 1:position + 2] in ['ISL', 'YSL']:
            self.next = (None, 1)
        # special case 'sugar-'
        elif (position == start_index
              and buffer[start_index:start_index + 5] == 'SUGAR'):
            self.next = ('X', 'S', 1)
        elif buffer[position:position + 2] == 'SH':
            # germanic
            if buffer[position + 1:position + 5] in [
                "HEIM", "HOEK", "HOLM", "HOLZ"]:
                self.next = ('S', 2)
            else:
                self.next = ('X', 2)
        # italian & armenian
        elif (buffer[position:position + 3] in ["SIO", "SIA"]
              or buffer[position:position + 4] == 'SIAN'):
            if not self.word.is_slavo_germanic:
                self.next = ('S', 'X', 3)
            else:
                self.next = ('S', 3)
        # german & anglicisations, e.g. 'smith' match 'schmidt', 'snider'
        # match 'schneider' also, -sz- in slavic language altho in
        # hungarian it is pronounced 's'
        elif ((position == start_index
               and buffer[position + 1] in ["M", "N", "L", "W"])
              or buffer[position + 1] == 'Z'):
            self.next = ('S', 'X')
            if buffer[position + 1] == 'Z':
                self.next = self.next + (2,)
            else:
                self.next = self.next + (1,)
        elif buffer[position:position + 2] == 'SC':
            # Schlesinger's rule
            if buffer[position + 2] == 'H':
                # dutch origin, e.g. 'school', 'schooner'
                if buffer[position + 3:position + 5] in [
                    "OO", "ER", "EN", "UY", "ED", "EM"]:
                    # 'schermerhorn', 'schenker'
                    if buffer[position + 3:position + 5] in ['ER', 'EN']:
                        self.next = ('X', 'SK', 3)
                    else:
                        self.next = ('SK', 3)
                else:
                    if (position == start_index
                        and buffer[start_index + 3] not in VOWELS
                        and buffer[start_index + 3] != 'W'):
                        self.next = ('X', 'S', 3)
                    else:
                        self.next = ('X', 3)
            elif buffer[position + 2] in ['I', 'E', 'Y']:
                self.next = ('S', 3)
            else:
                self.next = ('SK', 3)
        # french e.g. 'resnais', 'artois'
        elif (position == end_index
              and buffer[position - 2:position] in ['AI', 'OI']):
            self.next = ('', 'S', 1)
        else:
            self.next = ('S', )
            if buffer[position + 1] in ['S', 'Z']:
                self.next = self.next + (2,)
            else:
                self.next = self.next + (1,)

    def process_t(self):
        buffer = self.word.buffer
        position = self.position
        start_index = self.word.start_index
        if buffer[position:position + 4] == 'TION':
            self.next = ('X', 3)
        elif buffer[position:position + 3] in ['TIA', 'TCH']:
            self.next = ('X', 3)
        elif (buffer[position:position + 2] == 'TH'
              or buffer[position:position + 3] == 'TTH'):
            # special case 'thomas', 'thames' or germanic
            if (buffer[position + 2:position + 4] in ['OM', 'AM']
                or buffer[start_index:start_index + 4] in ['VON ', 'VAN ']
                or buffer[start_index:start_index + 3] == 'SCH'):
                self.next = ('T', 2)
            else:
                self.next = ('0', 'T', 2)
        elif buffer[position + 1] in ['T', 'D']:
            self.next = ('T', 2)
        else:
            self.next = ('T', 1)

    def process_v(self):
        if self.word.buffer[self.position + 1] == 'V':
            self.next = ('F', 2)
        else:
            self.next = ('F', 1)

    def process_w(self):
        buffer = self.word.buffer
        position = self.position
        start_index = self.word.start_index
        # can also be in middle of word
        if buffer[position:position + 2] == 'WR':
            self.next = ('R', 2)
        elif (position == start_index
            and (buffer[position + 1] in VOWELS
                 or buffer[position:position + 2] == 'WH')):
            # Wasserman should match Vasserman
            if buffer[position + 1] in VOWELS:
                self.next = ('A', 'F', 1)
            else:
                self.next = ('A', 1)
        # Arnow should match Arnoff
        elif ((position == self.word.end_index
               and buffer[position - 1] in VOWELS)
              or buffer[position - 1:position + 4] in [
                "EWSKI", "EWSKY", "OWSKI", "OWSKY"]
              or buffer[start_index:start_index + 3] == 'SCH'):
            self.next = ('', 'F', 1)
        # polish e.g. 'filipowicz'
        elif buffer[position:position + 4] in ["WICZ", "WITZ"]:
            self.next = ('TS', 'FX', 4)
        else:  # default is to skip it
            self.next = (None, 1)

    def process_x(self):
        buffer = self.word.buffer
        position = self.position
        # french e.g. breaux
        self.next = (None, )
        if not (
            position == self.word.end_index
            and (buffer[position - 3:position] in ["IAU", "EAU"]
                 or buffer[position - 2:position] in ['AU', 'OU'])):
            self.next = ('KS',)
        if buffer[position + 1] in ['C', 'X']:
            self.next = self.next + (2,)
        else:
            self.next = self.next + (1,)

    def process_z(self):
        # chinese pinyin e.g. 'zhao'
        if self.word.buffer[self.position + 1] == 'H':
            self.next = ('J', )
        elif (
            self.word.buffer[self.position + 1:self.position + 3] in [
                "ZO", "ZI", "ZA"]
            or (self.word.is_slavo_germanic
                and self.position > self.word.start_index
                and self.word.buffer[self.position - 1] != 'T')):
            self.next = ('S', 'TS')
        else:
            self.next = ('S', )
        if (self.word.buffer[self.position + 1] == 'Z'
            or self.word.buffer[self.position + 1] == 'H'):
            self.next = self.next + (2,)
        else:
            self.next = self.next + (1,)

    def parse(self, input):
        self.word = Word(input)
        self.position = self.word.start_index
        self.check_word_start()
        # loop through chars in word.buffer
        while self.position <= self.word.end_index:
            character = self.word.buffer[self.position]
            if character in VOWELS:
                self.process_initial_vowels()
            elif character == ' ':
                self.position += 1
                continue
            elif character == 'B':
                self.process_b()
            elif character == 'C':
                self.process_c()
            elif character == 'D':
                self.process_d()
            elif character == 'F':
                self.process_f()
            elif character == 'G':
                self.process_g()
            elif character == 'H':
                self.process_h()
            elif character == 'J':
                self.process_j()
            elif character == 'K':
                self.process_k()
            elif character == 'L':
                self.process_l()
            elif character == 'M':
                self.process_m()
            elif character == 'N':
                self.process_n()
            elif character == 'P':
                self.process_p()
            elif character == 'Q':
                self.process_q()
            elif character == 'R':
                self.process_r()
            elif character == 'S':
                self.process_s()
            elif character == 'T':
                self.process_t()
            elif character == 'V':
                self.process_v()
            elif character == 'W':
                self.process_w()
            elif character == 'X':
                self.process_x()
            elif character == 'Z':
                self.process_z()
            if len(self.next) == 2:
                if self.next[0]:
                    self.primary_phone += self.next[0]
                    self.secondary_phone += self.next[0]
                self.position += self.next[1]
            elif len(self.next) == 3:
                if self.next[0]:
                    self.primary_phone += self.next[0]
                if self.next[1]:
                    self.secondary_phone += self.next[1]
                self.position += self.next[2]
        if self.primary_phone == self.secondary_phone:
            self.secondary_phone = ""
        return (self.primary_phone, self.secondary_phone)


# backwards compatibility for the pre-OO implementation
def doublemetaphone(input):
    """
    Given an input string, return a 2-tuple of the double metaphone codes for
    the provided string. The second element of the tuple will be an empty
    string if it is identical to the first element.
    """
    return DoubleMetaphone().parse(input)


# for backwards compatibility for the old name of the function
dm = doublemetaphone



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import unicodedata


class Word(object):
    """
    """
    def __init__(self, input):
        self.original = input
        if isinstance(input, bytes):
            self.decoded = input.decode('utf-8', 'ignore')
        else:
            self.decoded = input
        self.decoded = self.decoded.replace('\xc7', "s")
        self.decoded = self.decoded.replace('\xe7', "s")
        self.normalized = ''.join(
            (c for c in unicodedata.normalize('NFD', self.decoded)
            if unicodedata.category(c) != 'Mn'))
        self.upper = self.normalized.upper()
        self.length = len(self.upper)
        self.prepad = "--"
        self.start_index = len(self.prepad)
        self.end_index = self.start_index + self.length - 1
        self.postpad = "------"
        # so we can index beyond the begining and end of the input string
        self.buffer = self.prepad + self.upper + self.postpad

    @property
    def is_slavo_germanic(self):
        return (
            self.upper.find('W') > -1
            or self.upper.find('K') > -1
            or self.upper.find('CZ') > -1
            or self.upper.find('WITZ') > -1)

    def get_letters(self, start=0, end=None):
        if not end:
            end = start + 1
        start = self.start_index + start
        end = self.start_index + end
        return self.buffer[start:end]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function
import vat
import pytest

# If you add to this list, pick people who:
#
#  (a) Publish their VAT number openly,
#  (b) Won't mind,
#  (c) Will still be around when the next person reads this.
#
# The first three numbers in this list should be from the UK, Spain and
# Germany respectively.  Why?  Because we know that the UK gives out details,
# that Spain doesn't but does fuzzy matching itself, and that Germany is
# broken.
test_numbers = [
    # These are all courtesy of Santander, who publish these details on their
    # various websites (sometimes it takes a moment searching, but they're all
    # public).
    ('GB466264724',
     { 'name': 'Santander UK plc',
        'street': 'Group Taxation, Customer Services Centre, Narborough',
        'postcode': 'LE19 0AL',
        'city': 'Leicester',
        'state': 'Leicestershire' }),
    ('ESA39000013',
     { 'name': 'Banco Santander, S.A.',
        'city': 'Santander',
        'postcode': '39004' }),
    ('DE120492390',
     { 'name': 'Santander Consumer Bank AG',
       'street': 'Santander-Platz 1',
       'city': 'Mönchengladbach',
       'postcode': '41061' }),
    ('ATU15350108',
     { 'name': 'Santander Consumer Bank GmbH',
       'street': 'Donau-City Straße 6',
       'city': 'Wien',
       'postcode': '1220' }),
    ('DK30733053',
     { 'name': ('Santander Consumer Bank, Filial '
                + 'af Santander Consumer Bank AS, Norge'),
       'street': 'Stamholmen 147 6',
       'city': 'Hvidovre',
       'postcode': '2650' }),
    ('IT05634190010',
     { 'name': 'Santander Consumer Bank Spa',
       'street': 'Via Nizza 262/26',
       'city': 'Torino',
       'postcode': '10126' }),
    ('PL5272046102',
     { 'name': 'Santander Consumer Bank S.A.',
       'street': 'ul. Strzegomska 42c',
       'city': 'Wrocław',
       'postcode': '53-611' }),
    ('PT503811483',
     { 'name': 'Banco Santander Consumer Portugal S.A.',
       'street': 'Rue Castilho, 2',
       'city': 'Lisboa',
       'postcode': '1269-073' }),
    ('SE516406033601',
     { 'name': 'Santander Consumer Bank AS Norge, Sverige Filial',
       'street': 'FE 302',
       'city': 'Stockholm',
       'postcode': '171 75' }),

    # The National Bank of Greece
    ('EL094014201',
     { 'name': 'Τράπεζα της Ελλάδος Α.Ε.',
       'street': 'Αιόλου 86',
       'city': 'Αθήνα',
       'postcode': '10232' }),
]

bad_info = { 'name': 'I am a great big rabbit' }
    
def test_local_fuzzy_vies():
    """Test a VIES response that requires *we* do the fuzzy matching."""
    try:
        valid, response = vat.check_details(test_numbers[0][0],
                                        test_numbers[0][1])
        assert valid == True

        valid, response = vat.check_details(test_numbers[0][0], bad_info)
        assert valid == False
    except vat.VIESHTTPException as e:
        if e.code >= 500 and e.code <= 599:
            pytest.skip('EU VIES server is malfunctioning, so skipping test')
        else:
            raise
        
def test_remote_fuzzy_vies():
    """Test a VIES response where the member state does the fuzzy matching."""
    try:
        valid, response = vat.check_details(test_numbers[1][0],
                                            test_numbers[1][1])
        assert valid == True

        valid, response = vat.check_details(test_numbers[1][0], bad_info)
        assert valid == False
    except vat.VIESHTTPException as e:
        if e.code >= 500 and e.code <= 599:
            pytest.skip('EU VIES server is malfunctioning, so skipping test')
        else:
            raise
        
def test_non_compliant_vies():
    """Test a VIES response where the member state is non-compliant and
    does not provide details *or* do fuzzy matching."""
    try:
        valid, response = vat.check_details(test_numbers[2][0], bad_info)
        assert valid is None
        assert response.trader_info['name'] == bad_info['name']
    except vat.VIESHTTPException as e:
        if e.code >= 500 and e.code <= 599:
            pytest.skip('EU VIES server is malfunctioning, so skipping test')
        else:
            raise
        
def test_all_vies():
    """Test all the numbers we have above, to make sure that we think they
    match."""
    for vat_number, details in test_numbers:
        try:
            valid, response = vat.check_details(vat_number, details)
        except vat.VIESHTTPException as e:
            if e.code >= 500 and e.code <= 599:
                pytest.skip('EU VIES server is malfunctioning, so skipping test')
            else:
                raise
        
        if vat_number.startswith('DE'):
            assert valid == None
        else:
            if valid != True:
                print('%s %r' % (vat_number, details))
                
            assert valid == True

            valid, response = vat.check_details(vat_number, bad_info)
            assert valid == False



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function
import vat
import pytest
from vat import vrws

def test_vrws():
    try:
        for ms in vat.member_states:
            rates = vrws.get_rates(ms.code)
            assert isinstance(rates, vrws.Rates)
            assert len(rates.types[vrws.STANDARD]) > 0
    except vat.VRWSHTTPException as e:
        if e.code >= 500 and e.code <= 599:
            pytest.skip('EU VRWS server is malfunctioning, so skipping test')
        else:
            raise
        



import zipfile
import pkg_resources
import datetime
import io
import re
import decimal
import bisect
import copy
from lxml import etree

from ... import MemberState

STANDARD_RATE = 'Standard'
REDUCED_RATE = 'Reduced'

TABLE = '{urn:oasis:names:tc:opendocument:xmlns:table:1.0}'
TEXT = '{urn:oasis:names:tc:opendocument:xmlns:text:1.0}'

def v(obj):
    wrapper = etree.Element(TEXT + 'p')
    wrapper.text = '%s' % obj
    return wrapper

class RepeatArray(object):
    def __init__(self):
        self.index = []
        self.items = []

    def copy_item(self, item):
        return copy.deepcopy(item)

    def set_repeat(self, item, repeat):
        pass
    
    def insert_before(self, item, new_item):
        pass

    def insert_after(self, item, new_item):
        pass

    def before_split(self):
        pass

    def after_split(self):
        pass
    
    def get(self, ndx, split=False):
        rndx = bisect.bisect_left(self.index, ndx, 0, len(self.index) - 2)
        item = self.items[rndx]
        if split:
            repeat = self.index[rndx + 1] - self.index[rndx]
            if repeat != 1:
                self.before_split()
                andx = self.index[rndx]
                before = andx - ndx
                after = self.index[rndx + 1] - ndx - 1
                self.set_repeat(item, 1)
                if before:
                    item_before = self.copy_item(item)
                    self.set_repeat(item_before, before)
                    self.items.insert(rndx, item_before)
                    rndx += 1
                    self.index.insert(rndx, ndx)
                    self.insert_before(item, item_before)
                if after:
                    item_after = self.copy_item(item)
                    self.set_repeat(item_after, after)
                    self.items.insert(rndx + 1, item_after)
                    self.index.insert(rndx + 1, ndx + 1)
                    self.insert_after(item, item_after)
                self.after_split()
        return item
    
class Row(RepeatArray):
    def __init__(self, row):
        super(Row, self).__init__()
        self.row = row

        colndx = 1
        for col in row.iterfind(TABLE + 'table-cell'):
            self.index.append(colndx)
            self.items.append(col)
            repeat = col.get(TABLE + 'number-columns-repeated')
            if repeat:
                colndx += int(repeat)
            else:
                colndx += 1
        self.index.append(colndx)

    def set_repeat(self, column, repeat):
        if repeat == 1:
            try:
                del column.attrib[TABLE + 'number-columns-repeated']
            except KeyError:
                pass
        else:
            column.set(TABLE + 'number-columns-repeated', '%s' % repeat)

    def insert_before(self, column, new_column):
        ndx = self.row.index(column)
        self.row.insert(ndx, new_column)

    def insert_after(self, column, new_column):
        ndx = self.row.index(column)
        self.row.insert(ndx + 1, new_column)

_ref_re = re.compile(r'([A-Z]+)([0-9]+)')
class Sheet(RepeatArray):
    def __init__(self, table):
        super(Sheet, self).__init__()
        self.table = table
        
        rowndx = 1
        for row in table.iterfind(TABLE + 'table-row'):
            self.index.append(rowndx)
            self.items.append(row)
            repeat = row.get(TABLE + 'number-rows-repeated')
            if repeat:
                rowndx += int(repeat)
            else:
                rowndx += 1
        self.index.append(rowndx)

    def set_repeat(self, row, repeat):
        if repeat == 1:
            try:
                del row.attrib[TABLE + 'number-rows-repeated']
            except KeyError:
                pass
        else:
            row.set(TABLE + 'number-rows-repeated', '%s' % repeat)

    def insert_before(self, row, new_row):
        ndx = self.table.index(row)
        self.table.insert(ndx, new_row)

    def insert_after(self, row, new_row):
        ndx = self.table.index(row)
        self.table.insert(ndx + 1, new_row)
        
    def cell(self, cell):
        m = _ref_re.match(cell.upper())
        row = int(m.group(2))
        col_a = m.group(1)
        col = 0
        for ch in col_a:
            chval = ord(ch) - ord('A') + 1
            col = 26 * col + chval
        return Row(self.get(row, split=True)).get(col, split=True)

def table(content, name):
    return content.find('.//{table}table[@{table}name="{name}"]'.format(table=TABLE, name=name))

class TooManySupplies(Exception):
    """Raised if you attempt to generate a MOSS return with more than 150
    supplies."""
    pass

def generate(file_obj,
             quarter=None, year=None,
             uk_supplies=[],
             fe_supplies=[]):
    """Generate an HMRC VAT MOSS return document.

       :param file_obj: A file-like object to which the VAT MOSS return should
         be written.
       :param int quarter: The quarter (1, 2, 3 or 4) for this return.
         Defaults to the previous quarter.
       :param int year: The year (e.g. 2015) for this return.
         Defaults to the current year.
       :param uk_supplies: An iterable yielding (MSC, rate type, rate, net value,
         vat due) tuples.  The MSC should be a two-character EU VAT country code
         or a MemberState object; the rate type should be one of

           :py:const:`STANDARD_RATE`
           :py:const:`REDUCED_RATE`

         The net value should be the value in Sterling.
       :param fe_supplies: An iterable yielding (VAT number, MSC, rate
         type, rate, net value, vat due) tuples.  The VAT number should
         include the two-character EU VAT country code; the MSC should be a
         two-character EU VAT country code or a MemberState object; the rate
         type should be one of

           :py:const:`STANDARD_RATE`
           :py:const:`REDUCED_RATE`

         The net value should be the value in Sterling.
        """

    today = datetime.date.today()
    
    if year is None:
        year = today.year

    if quarter is None:
        quarter = ((today.month - 4) // 3) % 4 + 1

    template_stream = pkg_resources.resource_stream('vat.gb.moss',
                                                    'resources/return-template.ods')
      
    with zipfile.ZipFile(template_stream, 'r') as template:
        with zipfile.ZipFile(file_obj, 'w') as output:
            # Copy everything apart from the worksheet pages (which we find
            # and parse)
            for item in template.infolist():
                data = template.read(item)
                if item.filename == 'content.xml':
                    content = etree.fromstring(data)
                    content_info = item
                    uk_sheet = Sheet(table(content, 'UK_SUPPLIES'))
                    fe_sheet = Sheet(table(content,
                                           'FIXED_ESTABLISHMENT_SUPPLIES'))
                else:
                    output.writestr(item, data)

            # Build the Quarter field
            quarter = 'Q%s/%s' % (quarter, year)

            # Set up for two dp
            two_dp = decimal.Decimal('.01')
            
            # Fill-in the UK supplies worksheet
            e10 = uk_sheet.cell('E10')
            e10[:] = [v(quarter)]

            made_supplies = False
            for ndx,supply in enumerate(uk_supplies):
                msc,rate_type,rate,net_value,vat_due = supply

                if ndx == 150:
                    raise TooManySupplies()
                
                if not isinstance(msc, MemberState):
                    msc = MemberState.by_code(msc)

                row_ndx = ndx + 17

                msc_cell = uk_sheet.cell('D%s' % row_ndx)
                msc_cell[:] = [v(msc.name)]
                rtype_cell = uk_sheet.cell('E%s' % row_ndx)
                rtype_cell[:] = [v(rate_type)]
                rate_cell = uk_sheet.cell('F%s' % row_ndx)
                rate_cell[:] = [v(decimal.Decimal(rate).quantize(two_dp))]
                net_cell = uk_sheet.cell('G%s' % row_ndx)
                net_cell[:] = [v(decimal.Decimal(net_value).quantize(two_dp))]
                vat_cell = uk_sheet.cell('H%s' % row_ndx)
                vat_cell[:] = [v(decimal.Decimal(vat_due).quantize(two_dp))]
                made_supplies = True

            if made_supplies:
                made_supplies = 'Yes'
            else:
                made_supplies = 'No'
                
            e12 = uk_sheet.cell('E12')
            e12[:] = [v(made_supplies)]
            
            # Fill-in the fixed establishment worksheet
            e11 = fe_sheet.cell('E11')
            e11[:] = [v(quarter)]

            made_supplies = False
            for ndx,supply in enumerate(fe_supplies):
                vat_number,msc,rate_type,rate,net_value,vat_due = supply

                if ndx == 150:
                    raise TooManySupplies()
                
                if not isinstance(msc, MemberState):
                    msc = MemberState.by_code(msc)

                row_ndx = ndx + 18

                vn_cell = fe_sheet.cell('D%s' % row_ndx)
                vn_cell[:] = [v(vat_number)]
                msc_cell = fe_sheet.cell('E%s' % row_ndx)
                msc_cell[:] = [v(msc.name)]
                rtype_cell = fe_sheet.cell('F%s' % row_ndx)
                rtype_cell[:] = [v(rate_type)]
                rate_cell = fe_sheet.cell('G%s' % row_ndx)
                rate_cell[:] = [v(decimal.Decimal(rate).quantize(two_dp))]
                net_cell = fe_sheet.cell('H%s' % row_ndx)
                net_cell[:] = [v(decimal.Decimal(net_value).quantize(two_dp))]
                vat_cell = fe_sheet.cell('I%s' % row_ndx)
                vat_cell[:] = [v(decimal.Decimal(vat_due).quantize(two_dp))]
                made_supplies = True

            if made_supplies:
                made_supplies = 'Yes'
            else:
                made_supplies = 'No'
                
            e13 = fe_sheet.cell('E13')
            e13[:] = [v(made_supplies)]

            # Write the worksheets
            output.writestr(content_info,
                            etree.tostring(content, encoding='utf8'))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import decimal
from decimal import Decimal as D
import re
import time
import datetime
import six
from six.moves import http_client
from lxml import etree

# Standard Rate types
STANDARD = 'Standard'
REDUCED = 'Reduced'

# Standard Categories
BROADCASTING = 'Broadcasting Services'
TELECOMS = 'Telecommunication Services'
ESERVICES = 'E-Services'

VRWS_HOST = str('ec.europa.eu')
VRWS_PATH = str('/taxation_customs/tic/services/VatRateWebService')

class VRWSException(Exception):
    pass

_error_re = re.compile(r'^VATRATE-ERR-([0-9]+) - (.*)$')
_date_re = re.compile(r'^([0-9]{2,})-([0-9]{2})-([0-9]{2})$')

class VRWSErrorException(VRWSException):
    def __init__(self, code, reason):
        self.code = code
        self.reason = reason

    def __repr__(self):
        return 'VRWSErrorException({code}, {reason})'.format(code=self.code,
                                                             reason=repr(self.reason))

    def __unicode__(self):
        return 'VATRATE-ERR-{code} - {reason}'.format(code=self.code,
                                                      reason=self.reason)

    def __str__(self):
        return str(self.__unicode__())

class VRWSSOAPException(VRWSException):
    def __init__(self, code, string, actor, detail):
        self.code = code
        self.string = string
        self.actor = actor
        self.detail = detail
        
    def __repr__(self):
        return 'VRWSSOAPException(%r, %r, %r, %r)' % (self.code,
                                                      self.string,
                                                      self.actor,
                                                      self.detail)

    def __unicode__(self):
        return '%s - %s (%s)\n%s' % (self.code, self.string, self.actor, self.detail)

    def __str__(self):
        return str(self.__unicode__())    

class VRWSHTTPException(VRWSException):
    def __init__(self, code, message):
        self.code = code
        self.message = message

    def __repr__(self):
        return 'VIESHTTPException(%r, %r)' % (self.code, self.message)

    def __unicode__(self):
        return '%s - %s' % (self.code, self.message)

    def __str__(self):
        return str(self.__unicode__())        

class Rate(object):
    """Represents an individual VAT rate."""
    def __init__(self, rate, application_date, detail=None):
        # The VAT rate, as a decimal percentage
        self.rate = rate

        # The date from which this rate applies
        self.application_date = application_date

        # Detail, if any; this might indicate a special rate (for instance)
        self.detail = detail

    def __repr__(self):
        return 'Rate(%r, %r, %r)' % (self.rate, self.application_date,
                                     self.detail)

    def __unicode__(self):
        if self.detail:
            return '%g%% (%s, as of %s)' % (self.rate, self.detail,
                                            self.application_date)
        else:
            return '%g%% (as of %s)' % (self.rate, self.application_date)

    def __str__(self):
        return str(self.__unicode__())
        
class Rates(object):
    def __init__(self, types, categories, regions):
        # A dictionary indexed by rate type
        self.types = types

        # A dictionary that contains rates organised by category
        self.categories = categories

        # A dictionary that contains any regional rates that may apply
        self.regions = regions
    
SOAP_NS = '{http://schemas.xmlsoap.org/soap/envelope/}'
VRWS_NS = '{urn:ec.europa.eu:taxud:tic:services:VatRateWebService:types}'
VRWS_NSM = '{urn:ec.europa.eu:taxud:tic:services:VatRateWebService}'

def format_date(date):
    return '{y:04d}-{m:02d}-{d:02d}'.format(y=date.year,
                                            m=date.month,
                                            d=date.day)

def send_message(message):
    message = message.encode('utf-8')

    headers = { b'Content-Type': b'text/xml',
                b'SOAPAction': b'urn:ec.europa.eu:taxud:tic:services:VatRateWebService' }

    tries = 0
    response = None
    while response is None or (tries < 5
                               and response.status >= 500
                               and response.status <= 599):
        if tries > 0:
            if response is not None:
                # Cope with badly behaved web service returning 500 for non-
                # server errors
                body = response.read()
                if body.startswith(b'<soap:'):
                    root = etree.fromstring(body)
                    fault = root.find('.//' + SOAP_NS + 'Fault')

                    faultcode = fault.find('./{*}faultcode').text
                    faultstring = fault.find('./{*}faultstring').text
                    faultactor = fault.find('./{*}faultactor')
                    if faultactor is not None:
                        faultactor = faultactor.text
                    detail = fault.find('./{*}detail')
                    if detail is not None:
                        detail = detail.text

                    m = _error_re.match(faultstring)
                    if m:
                        raise VRWSErrorException(int(m.group(1)), m.group(2))
                    
                    raise VRWSSOAPException(faultcode, faultstring,
                                            faultactor, detail)
        
            time.sleep(tries)
        tries += 1
        conn = http_client.HTTPConnection(VRWS_HOST)
        conn.request(str('POST'), VRWS_PATH, message, headers)
        response = conn.getresponse()

    if response.status != 200:
        raise VRWSHTTPException(response.status, response.reason)

    return response

def parse_response(response, kind):
    tree = etree.parse(response)
    root = tree.getroot()

    if root.tag.lower() != SOAP_NS + 'envelope':
        raise ValueError('Bad SOAP reply "%s"' % etree.tostring(tree))

    resp = root.find('./' + SOAP_NS + 'Body/' + VRWS_NSM + kind)

    if resp is None:
        raise ValueError('Bad SOAP reply "%s"' % etree.tostring(tree))

    types = {}
    categories = {}
    regions = {}
    
    for rate in resp.iter(VRWS_NS + 'rate'):
        rtype = rate.find('./' + VRWS_NS + 'type').text
        rvalue = D(rate.find('./' + VRWS_NS + 'value').text)
        rdate = rate.find('./' + VRWS_NS + 'applicationDate').text
        m = _date_re.match(rdate)
        rdate = datetime.date(int(m.group(1)),
                              int(m.group(2)),
                              int(m.group(3)))
        rrgn = rate.find('./' + VRWS_NS + 'region')
        if rrgn is not None:
            rrgn = rrgn.text
        rcat = rate.find('./' + VRWS_NS + 'category')
        if rcat is not None:
            rcat = rcat.text
        rdetail = rate.find('./' + VRWS_NS + 'detail')
        if rdetail is not None:
            rdetail = rdetail.text
            
        robj = Rate(rvalue, rdate, rdetail)
            
        if rrgn:
            rgn = regions.get(rrgn, None)
            if rgn is None:
                rgn = Rates({}, {}, None)
                regions[rrgn] = rgn
            if rcat:
                rgn.categories.setdefault(rcat, []).append(robj)
            else:
                rgn.types.setdefault(rtype, []).append(robj)
        else:
            if rcat:
                categories.setdefault(rcat, []).append(robj)
            else:
                types.setdefault(rtype, []).append(robj)

    return Rates(types, categories, regions)    

def get_rates(country, date=None,
              fetch_reduced=True, fetch_category=True, fetch_region=True):
    """Retrieve the VAT rates for the specified country.  Returns a
       Rates object on success, or in case of error raises an exception.

       """

    boolean = { True: 'true',
                False: 'false' }

    if date is None:
        date = datetime.date.today()

    fmtdate = format_date(date)

    message = '''<?xml version="1.0" encoding="UTF-8" ?>
<env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/"
 env:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
  <env:Body
   xmlns:vrwsm="urn:ec.europa.eu:taxud:tic:services:VatRateWebService"
   xmlns:vrws="urn:ec.europa.eu:taxud:tic:services:VatRateWebService:types">
    <vrwsm:ratesRequest>
      <vrws:memberState>{country}</vrws:memberState>
      <vrws:requestDate>{date}</vrws:requestDate>
      <vrws:fetchReduced>{fetch_reduced}</vrws:fetchReduced>
      <vrws:fetchCategory>{fetch_category}</vrws:fetchCategory>
      <vrws:fetchRegion>{fetch_region}</vrws:fetchRegion>
    </vrwsm:ratesRequest>
  </env:Body>
</env:Envelope>'''.format(country=country, date=fmtdate,
                          fetch_reduced=boolean[fetch_reduced],
                          fetch_category=boolean[fetch_category],
                          fetch_region=boolean[fetch_region])

    return parse_response(send_message(message), 'ratesResponse')

def get_changes(from_date=None, to_date=None, country=None):
    """Retrieve a list of VAT rate changes starting from `from_date`."""

    if from_date is None:
        from_date = datetime.date.today()
        
    extras = []
    
    if to_date is not None:
        extras.append('\n      <vrws:dateTo>{to_date}</vrws:dateTo>'\
                      .format(to_date=format_date(to_date)))

    if country is not None:
        extras.append('\n      <vrws:memberState>{country}</vrws:memberState>'\
                      .format(country=country))

    message = '''<?xml version="1.0" encoding="UTF-8" ?>
<env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/"
 env:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
 <env:Body
   xmlns:vrwsm="urn:ec.europa.eu:taxud:tic:services:VatRateWebService"
   xmlns:vrws="urn:ec.europa.eu:taxud:tic:services:VatRateWebService:types">
    <vrwsm:changesRequest>
      <vrws:dateFrom>{from_date}</vrws:dateFrom>{extras}
    </vrwsm:changesRequest>
  </env:Body>
</env:Envelope>'''.format(from_date=format_date(from_date),
                          extras=''.join(extras))

    return parse_response(send_message(message), 'changesResponse')
    
